/**
* Filename: StreamSerializer.java (in org.openbandy.io.stream)
* This file is part of the OpenBandy project.
* 
* www.openbandy.org
* 
*/

package org.openbandy.io.stream;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;

import org.openbandy.io.Serializable;
import org.openbandy.io.Serializer;

/**
 * The StreamSerializer is an extension of the <code>Serializer</code>
 * interface. It provides methods to serialize an object into a stream of
 * (compressed) XML. Thereby, different degrees of compression are possible.
*
* <br><br>
* (c) Copyright P. Bolliger 2007, ALL RIGHTS RESERVED.
*
* @author Philipp Bolliger (philipp@bolliger.name)
* @version 1.2
*/
public interface StreamSerializer extends Serializer {
    
    /**
     * The resulting stream will not be compressed.
     */
    public final static int COMPRESSION_NONE = 0;
    
    /**
     * The resulting stream will be compressed by replacing
     * common values and attributes with tokens.
     */
    public final static int COMPRESSED_TOKEN = 1;
    
    /**
     * The resulting stream will ge compressed using zip compression.
     */
    public final static int COMPRESSED_ZIP = 2;
    
    /**
     * The resulting stream will be compressed by replacing
     * common values and attributes with tokens and additionaly
     * compressed using zip.
     */
    public final static int COMPRESSED_BEST = 3;
    
    
    /**
     * This method generates an XML representation of the given object. If desired,
     * the resulting stream may be compressed.
     * 
     * @param serializable The object that needs to be serialized.
     * @param compression The degree of compression desired.
     * @return An OutputStream of serialized XML.
     */
    public ByteArrayOutputStream getXmlStream(Serializable serializable, int compression);
    
    /**
     * This method deserializes an object from the given InputStream. In doing
     * so, the degree of compression is resolved by this method.
     * 
     * @param inputStream
     *            The InputStream to read from.
     * 
     * @return The object that resulted from deserialization. In case of
     *         failure, it will return <code>NULL</code>.
     * 
     */
    public Serializable getObjectFromXmlStream(ByteArrayInputStream inputStream);

    /**
     * This method creates a new serializable object according
     * to the class name that can be read next from the inputstream
     * in the streamSerializationBuffer.
     * 
     * @param content Reference to the content object.
     * @param streamSerializationBuffer Proxy containing the inputstream.
     */
    public Serializable createContentObject(StreamSerializationBuffer streamSerializationBuffer);
}
